SegmentIterator

Struct SegmentIterator 

pub struct SegmentIterator<'a, M: AnyFrameMeta + Repr<MetaSlotStorage> + OwnerOf> { /* private fields */ }
Verified iterator over the frames owned by a Segment.

The iterator advances an owned suffix of the segment while threading the metadata region owner and segment owner permissions.

Implementations

Source

impl<'a, M: AnyFrameMeta + Repr<MetaSlotStorage> + OwnerOf> SegmentIterator<'a, M>

pub closed spec fn segment_ref(&self) -> &'a Segment<M>

pub closed spec fn range_spec(&self) -> Range<Paddr>

pub closed spec fn current_segment(&self) -> Segment<M>

pub closed spec fn remaining_spec(&self) -> Seq<SegmentIteratorItem<M>>

pub closed spec fn type_inv(self) -> bool

pub exec fn new(segment: &'a Segment<M>) -> res : Self

with
Tracked(regions): Tracked<&'a mut MetaRegionOwners>,
Tracked(owner): Tracked<&'a mut SegmentOwner<M>>,
requires
segment.invariants(owner, *regions),
ensures
res.segment_ref() == segment,
res.range_spec() == segment.range,
IteratorSpec::decrease(&res) is Some,
IteratorSpec::initial_value_relation(&res, &res),