pub struct SegmentOwner<M: AnyFrameMeta + ?Sized> {
pub perms: Seq<MetaPerm<M>>,
}Expand description
A SegmentOwner<M> holds the permission tokens for all frames in the
Segment<M> for verification purposes.
Fields
perms: Seq<MetaPerm<M>>The permissions for all frames in the segment, which must be well-formed and valid.